@@ -18,18 +18,13 @@
  */
 
 #include <signal.h>
-#include <errno.h>
-#include <string.h>
-#include <math.h>
 #include <unistd.h>
 
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 #include <cspi/spi.h>
-
-#include <libgnome/gnome-program.h>
-#include <libgnomeui/gnome-ui-init.h>
-#include <libgnomeui/gnome-client.h>
+#include <dbus/dbus-glib.h>
+#include <X11/extensions/XTest.h>
 
 #include "mt-common.h"
 #include "mt-service.h"
@@ -39,14 +34,95 @@
 #include "mt-cursor-manager.h"
 #include "mt-cursor.h"
 #include "mt-main.h"
+#include "mt-listener.h"
+#include "mt-accessible.h"
+
+#define GSM_DBUS_NAME      "org.gnome.SessionManager"
+#define GSM_DBUS_PATH      "/org/gnome/SessionManager"
+#define GSM_DBUS_INTERFACE "org.gnome.SessionManager"
+
+enum {
+    PRESS = 0,
+    RELEASE,
+    CLICK,
+    DOUBLE_CLICK
+};
+
+static GdkScreen *
+mt_main_current_screen (MTClosure *mt)
+{
+    GdkScreen *screen;
+
+    if (mt->n_screens > 1)
+	gdk_display_get_pointer (gdk_display_get_default (),
+				 &screen, NULL, NULL, NULL);
+    else
+	screen = gdk_screen_get_default ();
+
+    return screen;
+}
 
 static void
-mt_cursor_set (GdkCursorType type)
+mt_main_generate_motion_event (GdkScreen *screen, gint x, gint y)
 {
+    gdk_error_trap_push ();
+    gdk_display_warp_pointer (gdk_display_get_default (), screen, x, y);
+    gdk_flush ();
+    gdk_error_trap_pop ();
+}
+
+static void
+mt_main_generate_button_event (MTClosure *mt,
+			       guint      button,
+			       gboolean   type,
+			       gulong     delay)
+{
+    gdk_error_trap_push ();
+    switch (type) {
+	case PRESS:
+	    XTestFakeButtonEvent (mt->xtst_display,
+				  button, True, delay);
+	    break;
+	case RELEASE:
+	    XTestFakeButtonEvent (mt->xtst_display,
+				  button, False, delay);
+	    break;
+	case CLICK:
+	    XTestFakeButtonEvent (mt->xtst_display,
+				  button, True, CurrentTime);
+	    XTestFakeButtonEvent (mt->xtst_display,
+				  button, False, delay);
+	    break;
+	case DOUBLE_CLICK:
+	    XTestFakeButtonEvent (mt->xtst_display,
+				  button, True, CurrentTime);
+	    XTestFakeButtonEvent (mt->xtst_display,
+				  button, False, delay);
+	    XTestFakeButtonEvent (mt->xtst_display,
+				  button, True, CurrentTime);
+	    XTestFakeButtonEvent (mt->xtst_display,
+				  button, False, delay);
+	    break;
+	default:
+	    g_warning ("Unknown sequence.");
+	    break;
+    }
+    XFlush (mt->xtst_display);
+    gdk_error_trap_pop ();
+}
+
+static void
+mt_main_set_cursor (MTClosure *mt, GdkCursorType type)
+{
+    GdkScreen *screen;
     GdkCursor *cursor;
+    gint i;
 
     cursor = gdk_cursor_new (type);
-    gdk_window_set_cursor (gdk_get_default_root_window (), cursor);
+    for (i = 0; i < mt->n_screens; ++i) {
+	screen = gdk_display_get_screen (gdk_display_get_default (), i);
+	gdk_window_set_cursor (gdk_screen_get_root_window (screen), cursor);
+    }
     gdk_cursor_unref (cursor);
 }
 
@@ -59,54 +135,45 @@ dwell_restore_single_click (MTClosure *mt)
     mt_service_set_clicktype (mt->service, DWELL_CLICK_TYPE_SINGLE, NULL);
 }
 
-static gboolean
-dwell_b1r_timeout (gpointer data)
-{
-    gint x, y;
-
-    gdk_display_get_pointer (gdk_display_get_default (), NULL, &x, &y, NULL);
-    SPI_generateMouseEvent (x, y, "b1r");
-
-    return FALSE;
-}
-
 static void
-dwell_do_pointer_click (MTClosure *mt, gint x, gint y)
+mt_main_do_dwell_click (MTClosure *mt)
 {
     guint clicktype;
 
     clicktype = mt_service_get_clicktype (mt->service);
 
+    if (mt->dwell_mode == DWELL_MODE_GESTURE && !mt->dwell_drag_started)
+	mt_main_generate_motion_event (mt_main_current_screen (mt),
+				       mt->pointer_x, mt->pointer_y);
+
     switch (clicktype) {
-    case DWELL_CLICK_TYPE_SINGLE:
-	SPI_generateMouseEvent (x, y, "b1p");
-	/* Wait a few msecs before releasing the button again.
-	 * This allows GOK to see our clicks.
-	 */
-	g_timeout_add (60, dwell_b1r_timeout, NULL);
-	break;
-    case DWELL_CLICK_TYPE_DOUBLE:
-	SPI_generateMouseEvent (x, y, "b1d");
-	dwell_restore_single_click (mt);
-	break;
-    case DWELL_CLICK_TYPE_DRAG:
-	if (!mt->dwell_drag_started) {
-	    SPI_generateMouseEvent (x, y, "b1p");
-	    mt->dwell_drag_started = TRUE;
-	    mt_cursor_set (GDK_FLEUR);
-	}
-	else {
-	    SPI_generateMouseEvent (x, y, "b1r");
-	    mt->dwell_drag_started = FALSE;
-	    mt_cursor_set (GDK_LEFT_PTR);
+	case DWELL_CLICK_TYPE_SINGLE:
+	    mt_main_generate_button_event (mt, 1, CLICK, 60);
+	    break;
+	case DWELL_CLICK_TYPE_DOUBLE:
+	    mt_main_generate_button_event (mt, 1, DOUBLE_CLICK, 10);
 	    dwell_restore_single_click (mt);
-	}
-	break;
-    case DWELL_CLICK_TYPE_RIGHT:
-	SPI_generateMouseEvent (x, y, "b3c");
-	dwell_restore_single_click (mt);
-    default:
-	break;
+	    break;
+	case DWELL_CLICK_TYPE_DRAG:
+	    if (!mt->dwell_drag_started) {
+		mt_main_generate_button_event (mt, 1, PRESS, CurrentTime);
+		mt_main_set_cursor (mt, GDK_FLEUR);
+		mt->dwell_drag_started = TRUE;
+	    }
+	    else {
+		mt_main_generate_button_event (mt, 1, RELEASE, CurrentTime);
+		mt_main_set_cursor (mt, GDK_LEFT_PTR);
+		mt->dwell_drag_started = FALSE;
+		dwell_restore_single_click (mt);
+	    }
+	    break;
+	case DWELL_CLICK_TYPE_RIGHT:
+	    mt_main_generate_button_event (mt, 3, CLICK, 10);
+	    dwell_restore_single_click (mt);
+	    break;
+	default:
+	    g_warning ("Unknown click-type.");
+	    break;
     }
 }
 
@@ -122,16 +189,20 @@ below_threshold (MTClosure *mt, gint x, gint y)
 }
 
 static gboolean
-analyze_direction (MTClosure *mt, gint x, gint y)
+mt_main_analyze_gesture (MTClosure *mt)
 {
-    gint gd, i, dx, dy;
+    gint x, y, gd, i, dx, dy;
+
+    if (mt_service_get_clicktype (mt->service) == DWELL_CLICK_TYPE_DRAG)
+	return TRUE;
+
+    gdk_display_get_pointer (gdk_display_get_default (), NULL, &x, &y, NULL);
+    if (below_threshold (mt, x, y))
+	return FALSE;
 
     dx = ABS (x - mt->pointer_x);
     dy = ABS (y - mt->pointer_y);
 
-    if (below_threshold (mt, x, y))
-	return FALSE;
-
     /* find direction */
     if (x < mt->pointer_x)
 	if (y < mt->pointer_y)
@@ -157,25 +228,22 @@ analyze_direction (MTClosure *mt, gint x, gint y)
 		gd = DIRECTION_RIGHT;
 
     /* get click type for direction */
-    for (i = 0; i < N_CLICK_TYPES; i++)
+    for (i = 0; i < N_CLICK_TYPES; i++) {
 	if (mt->dwell_dirs[i] == gd) {
 	    mt_service_set_clicktype (mt->service, i, NULL);
 	    return TRUE;
 	}
-
+    }
     return FALSE;
 }
 
 static void
-draw_line (gint x1, gint y1, gint x2, gint y2)
+mt_main_draw_line (MTClosure *mt, gint x1, gint y1, gint x2, gint y2)
 {
-    GdkScreen *screen;
     GdkWindow *root;
     GdkGC *gc;
 
-    screen = gdk_display_get_default_screen (gdk_display_get_default ());
-    root = gdk_screen_get_root_window (screen);
-
+    root = gdk_screen_get_root_window (mt_main_current_screen (mt));
     gc = gdk_gc_new (root);
     gdk_gc_set_subwindow (gc, GDK_INCLUDE_INFERIORS);
     gdk_gc_set_function (gc, GDK_INVERT);
@@ -183,8 +251,7 @@ draw_line (gint x1, gint y1, gint x2, gint y2)
 				GDK_LINE_SOLID,
 				GDK_CAP_ROUND,
 				GDK_JOIN_ROUND);
-    gdk_draw_arc (root, gc, TRUE,
-		  x1 - 4, y1 - 4, 8, 8, 0, 23040);
+    gdk_draw_arc (root, gc, TRUE, x1 - 4, y1 - 4, 8, 8, 0, 23040);
     gdk_draw_line (root, gc, x1, y1, x2, y2);
     g_object_unref (gc);
 }
@@ -192,20 +259,21 @@ draw_line (gint x1, gint y1, gint x2, gint y2)
 static void
 dwell_start_gesture (MTClosure *mt)
 {
-    if (mt->override_cursor) {
-	GdkCursor *cursor;
+    GdkCursor *cursor;
+    GdkWindow *root;
 
+    if (mt->override_cursor) {
 	cursor = gdk_cursor_new (GDK_CROSS);
-	gdk_pointer_grab (gdk_get_default_root_window (),
-			  FALSE,
+	root = gdk_screen_get_root_window (mt_main_current_screen (mt));
+	gdk_pointer_grab (root, FALSE,
 			  GDK_POINTER_MOTION_MASK, 
-			  NULL,
-			  cursor,
-			  GDK_CURRENT_TIME);
+			  NULL, cursor,
+			  gtk_get_current_event_time ());
 	gdk_cursor_unref (cursor);
     }
-    else
-	mt_cursor_set (GDK_CROSS);
+    else {
+	mt_main_set_cursor (mt, GDK_CROSS);
+    }
 
     mt->dwell_gesture_started = TRUE;
     mt_timer_start (mt->dwell_timer);
@@ -215,12 +283,14 @@ static void
 dwell_stop_gesture (MTClosure *mt)
 {
     if (mt->override_cursor)
-	gdk_pointer_ungrab (GDK_CURRENT_TIME);
+	gdk_pointer_ungrab (gtk_get_current_event_time ());
     else
-	mt_cursor_set (GDK_LEFT_PTR);
+	mt_main_set_cursor (mt, GDK_LEFT_PTR);
 
     if (mt->x_old > -1 && mt->y_old > -1) {
-	draw_line (mt->pointer_x, mt->pointer_y, mt->x_old, mt->y_old);
+	mt_main_draw_line (mt, 
+			   mt->pointer_x, mt->pointer_y,
+			   mt->x_old, mt->y_old);
 	mt->x_old = -1;
 	mt->y_old = -1;
     }
@@ -233,30 +303,85 @@ static void
 dwell_timer_finished (MtTimer *timer, gpointer data)
 {
     MTClosure *mt = data;
-    gint x, y;
 
-    gdk_display_get_pointer (gdk_display_get_default (), NULL, &x, &y, NULL);
     mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
 
-    if (mt->dwell_mode == DWELL_MODE_CTW)
-	dwell_do_pointer_click (mt, x, y);
-    else if (mt->dwell_mode == DWELL_MODE_GESTURE) {
+    if (mt->dwell_mode == DWELL_MODE_CTW) {
+	mt_main_do_dwell_click (mt);
+    }
+    else {
 	if (mt->dwell_gesture_started) {
 	    dwell_stop_gesture (mt);
-	    if (analyze_direction (mt, x, y))
-		dwell_do_pointer_click (mt, mt->pointer_x, mt->pointer_y);
+
+	    if (mt_main_analyze_gesture (mt))
+		mt_main_do_dwell_click (mt);
+	}
+	/* if a drag action is in progress stop it */
+	else if (mt->dwell_drag_started) {
+	    mt_main_do_dwell_click (mt);
 	}
 	else
 	    dwell_start_gesture (mt);
     }
 }
 
+static gboolean
+eval_func (Accessible *a, gpointer data)
+{
+    gboolean found;
+    char *name;
+
+    name = Accessible_getName (a);
+    found = g_str_equal (name, "Window List");
+    SPI_freeString (name);
+
+    return found;
+}
+
+static gboolean 
+push_func (Accessible *a, gpointer data)
+{
+    MTClosure *mt = data;
+    AccessibleRole role;
+
+    role = Accessible_getRole (a);
+    if (role != SPI_ROLE_PANEL && role != SPI_ROLE_EMBEDDED)
+	return FALSE;
+
+    if (!mt_accessible_is_visible (a))
+	return FALSE;
+
+    if (Accessible_isComponent (a))
+	return mt_accessible_in_extents (a, mt->pointer_x, mt->pointer_y);
+
+    return TRUE;
+}
+
+static gboolean
+mt_main_use_move_release (MTClosure *mt)
+{
+    Accessible *point, *search;
+
+    point = mt_accessible_at_point (mt->pointer_x, mt->pointer_y);
+    if (point) {
+	search = mt_accessible_search (point,
+				       MT_SEARCH_TYPE_BREADTH,
+				       eval_func, push_func, mt);
+	Accessible_unref (point);
+	if (search) {
+	    Accessible_unref (search);
+	    return TRUE;
+	}
+    }
+    return FALSE;
+}
+
 static gboolean
 right_click_timeout (gpointer data)
 {
     MTClosure *mt = data;
 
-    SPI_generateMouseEvent (mt->pointer_x, mt->pointer_y, "b3c");
+    mt_main_generate_button_event (mt, 3, CLICK, CurrentTime);
 
     return FALSE;
 }
@@ -265,41 +390,60 @@ static void
 delay_timer_finished (MtTimer *timer, gpointer data)
 {
     MTClosure *mt = data;
+    GdkScreen *screen;
 
     mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
 
-    SPI_generateMouseEvent (0, 0, "b1r");
-    SPI_generateMouseEvent (mt->pointer_x, mt->pointer_y, "abs");
-
+    if (mt->move_release || mt_main_use_move_release (mt)) {
+	/* release the click outside of the focused object to
+	 * abort any action started by button-press.
+	 */
+	screen = mt_main_current_screen (mt);
+	mt_main_generate_motion_event (screen, 0, 0);
+	mt_main_generate_button_event (mt, 1, RELEASE, CurrentTime);
+	mt_main_generate_motion_event (screen, mt->pointer_x, mt->pointer_y);
+    }
+    else {
+	mt_main_generate_button_event (mt, 1, RELEASE, CurrentTime);
+    }
+    /* wait 100 msec before releasing the button again -
+     * gives apps some time to release active grabs, eg: gnome-panel 'move'
+     */
     g_timeout_add (100, right_click_timeout, data);
 }
 
-/* at-spi callbacks */
+/* at-spi listener callbacks */
 static void
-spi_motion_event (const AccessibleEvent *event, void *data)
+global_motion_event (MtListener *listener,
+		     MtEvent    *event,
+		     gpointer    data)
 {
     MTClosure *mt = data;
 
     if (mt->dwell_enabled) {
-	if (!below_threshold (mt, event->detail1, event->detail2) &&
+	if (!below_threshold (mt, event->x, event->y) &&
 	    !mt->dwell_gesture_started) {
-	    mt->pointer_x = (gint) event->detail1;
-	    mt->pointer_y = (gint) event->detail2;
+	    mt->pointer_x = event->x;
+	    mt->pointer_y = event->y;
 	    mt_timer_start (mt->dwell_timer);
 	}
+
 	if (mt->dwell_gesture_started) {
-	    if (mt->x_old > -1 && mt->y_old > -1)
-		draw_line (mt->pointer_x, mt->pointer_y, mt->x_old, mt->y_old);
-
-	    draw_line (mt->pointer_x, mt->pointer_y,
-		       event->detail1, event->detail2);
-
-	    mt->x_old = (gint) event->detail1;
-	    mt->y_old = (gint) event->detail2;
+	    if (mt->x_old > -1 && mt->y_old > -1) {
+		mt_main_draw_line (mt,
+				   mt->pointer_x, mt->pointer_y,
+				   mt->x_old, mt->y_old);
+	    }
+	    mt_main_draw_line (mt, 
+			       mt->pointer_x, mt->pointer_y,
+			       event->x, event->y);
+	    mt->x_old = event->x;
+	    mt->y_old = event->y;
 	}
     }
+
     if (mt_timer_is_running (mt->delay_timer)) {
-	if (!below_threshold (mt, event->detail1, event->detail2)) {
+	if (!below_threshold (mt, event->x, event->y)) {
 	    mt_timer_stop (mt->delay_timer);
 	    mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
 	}
@@ -307,20 +451,22 @@ spi_motion_event (const AccessibleEvent *event, void *data)
 }
 
 static void
-spi_button_event (const AccessibleEvent *event, void *data)
+global_button_event (MtListener *listener,
+		     MtEvent    *event,
+		     gpointer    data)
 {
     MTClosure *mt = data;
 
-    if (g_str_equal (event->type, "mouse:button:1p")) {
+    if (event->type == EV_BUTTON_PRESS) {
 	if (mt->delay_enabled) {
-	    mt->pointer_x = (gint) event->detail1;
-	    mt->pointer_y = (gint) event->detail2;
+	    mt->pointer_x = event->x;
+	    mt->pointer_y = event->y;
 	    mt_timer_start (mt->delay_timer);
 	}
 	if (mt->dwell_gesture_started)
 	    dwell_stop_gesture (mt);
     }
-    else if (g_str_equal (event->type, "mouse:button:1r")) {
+    else {
 	if (mt->delay_enabled) {
 	    mt_timer_stop (mt->delay_timer);
 	    mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
@@ -328,6 +474,21 @@ spi_button_event (const AccessibleEvent *event, void *data)
     }
 }
 
+static void
+global_focus_event (MtListener *listener, gpointer data)
+{
+    MTClosure *mt = data;
+    Accessible *accessible;
+
+    if (mt->delay_enabled) {
+	accessible = mt_listener_current_focus (listener);
+	/* TODO: check for more objects and conditions.
+	 * Some links don't have jump actions, eg: text-mails in thunderbird.
+	 */
+	mt->move_release = mt_accessible_supports_action (accessible, "jump");
+    }
+}
+
 static gboolean
 cursor_overlay_time (guchar  *image,
 		     gint     width,
@@ -335,6 +496,8 @@ cursor_overlay_time (guchar  *image,
 		     MtTimer *timer,
 		     gdouble  time)
 {
+    GtkWidget *ctw;
+    GdkColor c;
     cairo_surface_t *surface;
     cairo_t *cr;
     gdouble target;
@@ -352,10 +515,17 @@ cursor_overlay_time (guchar  *image,
 	return FALSE;
     }
 
+    ctw = mt_ctw_get_window ();
+    c = ctw->style->bg[GTK_STATE_SELECTED];
     target = mt_timer_get_target (timer);
+
     cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);
     cairo_rectangle (cr, 0, 0, width, height / (target / time));
-    cairo_set_source_rgba (cr, 0., 0., 0., 0.75);
+    cairo_set_source_rgba (cr,
+			   c.red   / 65535.,
+			   c.green / 65535.,
+			   c.blue  / 65535.,
+			   0.60);
     cairo_fill (cr);
     cairo_destroy (cr);
     cairo_surface_destroy (surface);
@@ -364,7 +534,9 @@ cursor_overlay_time (guchar  *image,
 }
 
 static void
-mt_update_cursor (MtCursor *cursor, MtTimer *timer, gdouble time)
+mt_main_update_cursor (MtCursor *cursor,
+		       MtTimer  *timer,
+		       gdouble   time)
 {
     guchar *image;
     gushort width, height;
@@ -393,21 +565,14 @@ mt_update_cursor (MtCursor *cursor, MtTimer *timer, gdouble time)
 }
 
 static void
-delay_timer_tick (MtTimer *timer, gdouble time, gpointer data)
+mt_main_timer_tick (MtTimer *timer,
+		    gdouble  time,
+		    gpointer data)
 {
     MTClosure *mt = data;
 
     if (mt->animate_cursor && mt->cursor != NULL)
-	mt_update_cursor (mt->cursor, timer, time);
-}
-
-static void
-dwell_timer_tick (MtTimer *timer, gdouble time, gpointer data)
-{
-    MTClosure *mt = data;
-
-    if (mt->animate_cursor && mt->cursor != NULL)
-	mt_update_cursor (mt->cursor, timer, time);
+	mt_main_update_cursor (mt->cursor, timer, time);
 }
 
 static void
@@ -435,7 +600,7 @@ cursor_changed (MtCursorManager *manager,
 static void
 signal_handler (int sig)
 {
-    SPI_event_quit ();
+    gtk_main_quit ();
 }
 
 static void
@@ -520,6 +685,29 @@ get_gconf_options (MTClosure *mt)
 	gconf_client_get_int (mt->client, OPT_G_RIGHT, NULL);
 }
 
+static void
+mt_main_request_logout (MTClosure *mt)
+{
+    DBusGConnection *bus;
+    DBusGProxy *proxy;
+
+    bus = dbus_g_bus_get (DBUS_BUS_SESSION, NULL);
+    if (bus) {
+	proxy = dbus_g_proxy_new_for_name (bus,
+					   GSM_DBUS_NAME,
+					   GSM_DBUS_PATH,
+					   GSM_DBUS_INTERFACE);
+	/*
+	 * Call logout method of session manager:
+	 * mode: 0 = normal, 1 = no confirmation, 2 = force
+	 */
+	dbus_g_proxy_call (proxy, "Logout", NULL,
+			   G_TYPE_UINT, 1, G_TYPE_INVALID,
+			   G_TYPE_INVALID);
+	g_object_unref (proxy);
+    }
+}
+
 static gboolean
 accessibility_enabled (MTClosure *mt,
 		       gint       spi_status)
@@ -538,13 +726,10 @@ accessibility_enabled (MTClosure *mt,
 	       "To enable support for assistive technologies and restart "
 	       "your session, press \"Enable and Log Out\"."),
 	     MT_MESSAGE_LOGOUT);
-	if (ret == GTK_RESPONSE_ACCEPT) {
-	    GnomeClient *session;
 
+	if (ret == GTK_RESPONSE_ACCEPT) {
 	    gconf_client_set_bool (mt->client, GNOME_A11Y_KEY, TRUE, NULL);
-	    if ((session = gnome_master_client ()))
-		gnome_client_request_save (session, GNOME_SAVE_GLOBAL, TRUE,
-					   GNOME_INTERACT_ANY, FALSE, TRUE);
+	    mt_main_request_logout (mt);
 	}
 	else {
 	    /* reset the selected option again */
@@ -562,10 +747,18 @@ static MTClosure *
 mt_closure_init (void)
 {
     MTClosure *mt;
+    gint ev_base, err_base, maj, min;
 
     mt = g_slice_new0 (MTClosure);
-    if (!mt)
+    mt->xtst_display = XOpenDisplay (NULL);
+
+    if (!XTestQueryExtension (mt->xtst_display,
+			      &ev_base, &err_base, &maj, &min)) {
+	XCloseDisplay (mt->xtst_display);
+	g_slice_free (MTClosure, mt);
+	g_critical ("No XTest extension found. Aborting..");
 	return NULL;
+    }
 
     mt->client = gconf_client_get_default ();
     gconf_client_add_dir (mt->client, MT_GCONF_HOME,
@@ -577,17 +770,19 @@ mt_closure_init (void)
     g_signal_connect (mt->delay_timer, "finished",
 		      G_CALLBACK (delay_timer_finished), mt);
     g_signal_connect (mt->delay_timer, "tick",
-		      G_CALLBACK (delay_timer_tick), mt);
+		      G_CALLBACK (mt_main_timer_tick), mt);
 
     mt->dwell_timer = mt_timer_new ();
     g_signal_connect (mt->dwell_timer, "finished",
 		      G_CALLBACK (dwell_timer_finished), mt);
     g_signal_connect (mt->dwell_timer, "tick",
-		      G_CALLBACK (dwell_timer_tick), mt);
+		      G_CALLBACK (mt_main_timer_tick), mt);
 
     mt->service = mt_service_get_default ();
     mt_service_set_clicktype (mt->service, DWELL_CLICK_TYPE_SINGLE, NULL);
 
+    mt->n_screens = gdk_display_get_n_screens (gdk_display_get_default ());
+
     mt->x_old = -1;
     mt->y_old = -1;
 
@@ -673,7 +868,7 @@ main (int argc, char **argv)
     g_print ("Starting daemon.\n");
 
     if ((pid = fork ()) < 0) {
-	g_print ("Fork failed.\n");
+	g_error ("fork() failed.");
 	return 1;
     }
     else if (pid) {
@@ -682,12 +877,11 @@ main (int argc, char **argv)
     }
     else {
 	/* Child process */
-	GnomeProgram *program;
 	MTClosure *mt;
 	MtCursorManager *manager;
-	AccessibleEventListener *bl, *ml;
+	MtListener *listener;
 	gint spi_status;
-	gint spi_leaks;
+	gint spi_leaks = 0;
 
 	if (mt_pidfile_create () < 0)
 	    return 1;
@@ -697,28 +891,19 @@ main (int argc, char **argv)
 	signal (SIGQUIT, signal_handler);
 	signal (SIGHUP, signal_handler);
 
+	gtk_init (&argc, &argv);
 	g_set_application_name ("Mousetweaks");
-	program = gnome_program_init (PACKAGE, VERSION, LIBGNOMEUI_MODULE,
-				      argc, argv, GNOME_PARAM_NONE);
 
 	mt = mt_closure_init ();
 	if (!mt)
 	    goto FINISH;
 
-	spi_leaks = 0;
 	spi_status = SPI_init ();
 	if (!accessibility_enabled (mt, spi_status)) {
 	    mt_closure_free (mt);
 	    goto FINISH;
 	}
 
-	/* add at-spi listeners */
-	ml = SPI_createAccessibleEventListener (spi_motion_event, (void *) mt);
-	SPI_registerGlobalEventListener (ml, "mouse:abs");
-	bl = SPI_createAccessibleEventListener (spi_button_event, (void *) mt);
-	SPI_registerGlobalEventListener (bl, "mouse:button:1p");
-	SPI_registerGlobalEventListener (bl, "mouse:button:1r");
-
 	/* command-line options */
 	if (dwell_click)
 	    gconf_client_set_bool (mt->client, OPT_DWELL, TRUE, NULL);
@@ -749,6 +934,7 @@ main (int argc, char **argv)
 	if (!mt_ctw_init (mt, pos_x, pos_y))
 	    goto CLEANUP;
 
+	/* init cursor animation */
 	manager = mt_cursor_manager_get_default ();
 	g_signal_connect (manager, "cursor_changed",
 			  G_CALLBACK (cursor_changed), mt);
@@ -757,24 +943,30 @@ main (int argc, char **argv)
 
 	mt->cursor = mt_cursor_manager_current_cursor (manager);
 
-	SPI_event_main ();
+	/* init at-spi signals */
+	listener = mt_listener_get_default ();
+	g_signal_connect (listener, "motion_event",
+			  G_CALLBACK (global_motion_event), mt);
+	g_signal_connect (listener, "button_event",
+			  G_CALLBACK (global_button_event), mt);
+	g_signal_connect (listener, "focus_changed",
+			  G_CALLBACK (global_focus_event), mt);
+
+	gtk_main ();
 
 	mt_cursor_manager_restore_all (manager);
 	g_object_unref (manager);
+	g_object_unref (listener);
 
 	CLEANUP:
-	    SPI_deregisterGlobalEventListenerAll (bl);
-	    AccessibleEventListener_unref (bl);
-	    SPI_deregisterGlobalEventListenerAll (ml);
-	    AccessibleEventListener_unref (ml);
 	    spi_leaks = SPI_exit ();
 	    mt_closure_free (mt);
 	FINISH:
 	    mt_pidfile_remove ();
-	    g_object_unref (program);
 
 	if (spi_leaks)
-	    g_warning ("AT-SPI reported %i leaks", spi_leaks);
+	    g_warning ("AT-SPI reported %i leak%s.",
+		       spi_leaks, spi_leaks != 1 ? "s" : "");
     }
     return 0;
 }
